<?php
/**
 * CSExcludeLogRoute class file.
 *
 * @author Maciej Lizewski (redguy)
 */

/**
 * CSExcludeLogRoute provides fine-grained, nested negative filters for logs. It works as a filter and send logs matching criterie to nested LogRoutes.
 *
 * TODO: provide support for nested profile log routes which use CLogger::getProfilingResults interface
 *
 * Just use in config files like this:
 *
 * 	'components'=>array(
 *		'log'=>array(
 *			'class'=>'CLogRouter',
 *			'routes'=>array(
 *				//log all logs from every category with every level except logs from 'system.db.*' and 'system.CModule' categories (any level)
 *				array(
 *					'class'=>'CSExcludeLogRoute',
 *					'excludeCategories'=>'system.db.*,system.CModule',
 *					'routes'=>array(
 *						array(
 *							'class'=>'CWebLogRoute',
 *						),
 *					),
 *				),
 *				//log all logs from system.* categories except logs from system.db.* categories which have 'warning' level
 *				array(
 *					'class'=>'CSExcludeLogRoute',
 *					'categories'=>'system.*',
 *					'excludeCategories'=>'system.db.*',
 *					'excludeLevels'=>'warning',
 *					'routes'=>array(
 *						array(
 *							'class'=>'CWebLogRoute',
 *						),
 *					),
 *				),
 *			),
 *		),
 *		...
 *	),
 *
 */
class CSExcludeLogRoute extends CLogRoute {

	/**
	 * @var string list of categories to be excluded separated by comma or space. Defaults to empty, meaning no categories will be excluded.
	 */
	public $excludeCategories;
	/**
	 * @var string list of levels to be excluded separated by comma or space. Defaults to empty, meaning no levels will be excluded.
	 */
	public $excludeLevels;
	
	
	private $_routes = array();
	private $_categories = array();
	private $_levels = array();
	private $_excludeCategories = array();
	private $_excludeLevels = array();
	private $_logs = array();

	/**
	 * Initializes the route.
	 * This method is invoked after the route is created by the route manager.
	 */
	public function init() {
		parent::init();
		foreach( $this->_routes as $name=>$route ) {
			$route = Yii::createComponent( $route );
			$route->init();
			$this->_routes[$name] = $route;
		}
	}

	/**
	 * @return array the currently initialized routes
	 */
	public function getRoutes() {
		return new CMap( $this->_routes );
	}

	/**
	 * @param array $config list of route configurations. Each array element represents
	 * the configuration for a single route and has the following array structure:
	 * <ul>
	 * <li>class: specifies the class name or alias for the route class.</li>
	 * <li>name-value pairs: configure the initial property values of the route.</li>
	 * </ul>
	 */
	public function setRoutes( $config ) {
		foreach( $config as $name=>$route ) {
			$this->_routes[$name] = $route;
		}
	}
	
	/**
	 * Retrieves log messages. This routine emulates CLogger interface for nested LogRoutes.
	 *
	 * @param string $levels level filter
	 * @param string $categories category filter
	 * @return array list of messages. Each array element represents one message
	 */
	public function getLogs( $levels = '', $categories = '' ) {
		$this->_levels = preg_split( '/[\s,]+/', strtolower( $levels ), -1, PREG_SPLIT_NO_EMPTY );
		$this->_categories = preg_split( '/[\s,]+/', strtolower( $categories ), -1, PREG_SPLIT_NO_EMPTY );
		
		if( empty( $levels ) && empty( $categories ) ) {
			return $this->_logs;
		} elseif( empty( $levels ) ) {
			return array_values( array_filter( $this->_logs, array( $this, 'filterByCategory' ) ) );
		} elseif( empty( $categories ) ) {
			return array_values( array_filter( $this->_logs, array( $this, 'filterByLevel' ) ) );
		} else {
			$ret = array_filter( $this->_logs, array( $this, 'filterByLevel' ) );
			return array_values( array_filter( $ret, array( $this, 'filterByCategory' ) ) );
		}
	}

	/**
	 * Retrieves filtered log messages from logger for further processing in nested LogRoutes.
	 * @param CLogger $logger logger instance
	 * @param boolean $processLogs whether to process the logs after they are collected from the logger
	 */
	public function collectLogs( $logger, $processLogs = false ) {
		$logs = $logger->getLogs( $this->levels, $this->categories );
		if( $logs === array() ) {
			$this->_logs = $logs;
		} else {
			$this->_excludeLevels = preg_split( '/[\s,]+/', strtolower( $this->excludeLevels ), -1, PREG_SPLIT_NO_EMPTY );
			$this->_excludeCategories = preg_split( '/[\s,]+/', strtolower( $this->excludeCategories ), -1, PREG_SPLIT_NO_EMPTY );

			if( !empty( $this->_excludeCategories ) ) {
				$this->_logs = array_values( array_filter( $logs, array( $this, 'filterExcludeByCategory' ) ) );
			} elseif( !empty( $this->_excludeLevels ) ) {
				$this->_logs = array_values( array_filter( $logs, array( $this, 'filterExcludeByLevel' ) ) );
			} else {
				$this->_logs = $logs;
			}
		}
		
		foreach( $this->_routes as $route ) {
			if( $route->enabled ) {
				$route->collectLogs( $this, $processLogs );
			}
		}
	}

	/**
	 * Saves log messages in files.
	 * @param array $logs list of log messages
	 */
	protected function processLogs( $logs ) {
	}

	/**
	 * Filter function used by {@link getLogs}
	 * @param array $value element to be filtered
	 * @return bool true if valid log, false if not.
	 */
	private function filterByLevel( $value ) {
		return in_array( strtolower( $value[1] ), $this->_levels );
	}
	
	/**
	 * Filter function used by {@link getLogs}
	 * @param array $value element to be filtered
	 * @return bool true if valid log, false if not.
	 */
	private function filterByCategory( $value ) {
		foreach( $this->_categories as $category ) {
			$cat = strtolower( $value[2] );
			if( $cat === $category || ( ( $c = rtrim( $category, '.*' ) ) !== $category && strpos( $cat, $c ) === 0 ) ) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Negative filter function used by {@link collectLogs}
	 * @param array $value element to be filtered
	 * @return bool true if valid log, false if not.
	 */
	private function filterExcludeByLevel( $value ) {
		return !empty( $this->_excludeLevels ) && !in_array( strtolower( $value[1] ), $this->_excludeLevels );
	}
	
	/**
	 * Negative filter function used by {@link collectLogs}
	 * @param array $value element to be filtered
	 * @return bool true if valid log, false if not.
	 */
	private function filterExcludeByCategory( $value ) {
		foreach( $this->_excludeCategories as $category ) {
			$cat = strtolower( $value[2] );
			if( $cat === $category || ( ( $c = rtrim( $category, '.*' ) ) !== $category && strpos( $cat, $c ) === 0 ) ) {
				return $this->filterExcludeByLevel( $value );
			}
		}
		return true;
	}
}
